module M_match
  implicit none
  private

  public :: glob
  public :: string_match
  public :: match_wild
  public :: Wildcmp

contains
!===================================================================================================================================
!()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()!
!===================================================================================================================================
!>
!!##NAME
!!    glob(3f) - [M_strings:COMPARE] compare given string for match to
!!    a pattern which may contain globbing wildcard characters
!!    (LICENSE:PD)
!!
!!##SYNOPSIS
!!
!!    logical function glob(string, pattern )
!!
!!     character(len=*),intent(in) :: string
!!     character(len=*),intent(in) :: pattern
!!
!!##DESCRIPTION
!!    glob(3f) compares given STRING for match to PATTERN which may
!!    contain basic wildcard "globbing" characters.
!!
!!    In this version to get a match the entire string must be described
!!    by PATTERN. Trailing whitespace is significant, so trim the input
!!    string to have trailing whitespace ignored.
!!
!!##OPTIONS
!!    string   the input string to test to see if it contains the pattern.
!!    pattern  the following simple globbing options are available
!!
!!             o "?" matching any one character
!!             o "*" matching zero or more characters.
!!               Do NOT use adjacent asterisks.
!!             o spaces are significant and must be matched or pretrimmed
!!             o There is no escape character, so matching strings with
!!               literal question mark and asterisk is problematic.
!!
!!##AUTHOR
!!   John S. Urban
!!
!!##REFERENCE
!!   The article "Matching Wildcards: An Empirical Way to Tame an Algorithm"
!!   in Dr Dobb's Journal, By Kirk J. Krauss, October 07, 2014
!!
!!##LICENSE
!!   Public Domain
function glob(tame,wild)

! ident_6="@(#)M_strings::glob(3f): function compares text strings, one of which can have wildcards ('*' or '?')."

logical                    :: glob
character(len=*)           :: tame       ! A string without wildcards
character(len=*)           :: wild       ! A (potentially) corresponding string with wildcards
character(len=len(tame)+1) :: tametext
character(len=len(wild)+1) :: wildtext
character(len=1),parameter :: NULL=char(0)
integer                    :: wlen
integer                    :: ti, wi
integer                    :: i
character(len=:),allocatable :: tbookmark, wbookmark
! These two values are set when we observe a wildcard character. They
! represent the locations, in the two strings, from which we start once we have observed it.
   tametext=tame//NULL
   wildtext=wild//NULL
   tbookmark = NULL
   wbookmark = NULL
   wlen=len(wild)
   wi=1
   ti=1
   do                                            ! Walk the text strings one character at a time.
      if(wildtext(wi:wi) == '*')then             ! How do you match a unique text string?
         do i=wi,wlen                            ! Easy: unique up on it!
            if(wildtext(wi:wi).eq.'*')then
               wi=wi+1
            else
               exit
            endif
         enddo
         if(wildtext(wi:wi).eq.NULL) then        ! "x" matches "*"
            glob=.true.
            return
         endif
         if(wildtext(wi:wi) .ne. '?') then
            ! Fast-forward to next possible match.
            do while (tametext(ti:ti) .ne. wildtext(wi:wi))
               ti=ti+1
               if (tametext(ti:ti).eq.NULL)then
                  glob=.false.
                  return                         ! "x" doesn't match "*y*"
               endif
            enddo
         endif
         wbookmark = wildtext(wi:)
         tbookmark = tametext(ti:)
      elseif(tametext(ti:ti) .ne. wildtext(wi:wi) .and. wildtext(wi:wi) .ne. '?') then
         ! Got a non-match. If we've set our bookmarks, back up to one or both of them and retry.
         if(wbookmark.ne.NULL) then
            if(wildtext(wi:).ne. wbookmark) then
               wildtext = wbookmark
               wlen=len_trim(wbookmark)
               wi=1
               ! Don't go this far back again.
               if (tametext(ti:ti) .ne. wildtext(wi:wi)) then
                  tbookmark=tbookmark(2:)
                  tametext = tbookmark
                  ti=1
                  cycle                          ! "xy" matches "*y"
               else
                  wi=wi+1
               endif
            endif
            if (tametext(ti:ti).ne.NULL) then
               ti=ti+1
               cycle                             ! "mississippi" matches "*sip*"
            endif
         endif
         glob=.false.
         return                                  ! "xy" doesn't match "x"
      endif
      ti=ti+1
      wi=wi+1
      if(ti.gt.len(tametext))then
         glob=.false.
         return
      elseif (tametext(ti:ti).eq.NULL) then          ! How do you match a tame text string?
         if(wildtext(wi:wi).ne.NULL)then
            do while (wildtext(wi:wi) == '*')    ! The tame way: unique up on it!
               wi=wi+1                           ! "x" matches "x*"
               if(wildtext(wi:wi).eq.NULL)exit
            enddo
         endif
         if (wildtext(wi:wi).eq.NULL)then
            glob=.true.
            return                               ! "x" matches "x"
         endif
         glob=.false.
         return                                  ! "x" doesn't match "xy"
      endif
   enddo
end function glob

recursive function string_match( string, pattern ) result(match)
! from an old version of the source code for FLIBS. 
implicit none

character(len=1), parameter :: backslash = '\'
character(len=1), parameter :: star = '*'
character(len=1), parameter :: question = '?'
character(len=*), intent(in) :: string
character(len=*), intent(in) :: pattern
logical :: match

character(len=len(pattern)) :: literal
integer :: ptrim
integer :: p
integer :: k
integer :: ll
integer :: method
integer :: start
integer :: strim

match = .false.
method = 0
ptrim = len_trim( pattern )
strim = len_trim( string )
p = 1
ll = 0
start = 1

!
! Split off a piece of the pattern
!
do while ( p <= ptrim )
   select case ( pattern(p:p) )
   case( star )
    if ( ll .ne. 0 ) exit
    method = 1
   case( question )
    if ( ll .ne. 0 ) exit
    method = 2
    start = start + 1
   case( backslash )
    p = p + 1
    ll = ll + 1
    literal(ll:ll) = pattern(p:p)
   case default
    ll = ll + 1
    literal(ll:ll) = pattern(p:p)
   end select
   
   p = p + 1
enddo

!
! Now look for the literal string (if any!)
!
if ( method == 0 ) then
   !
   ! We are at the end of the pattern, and of the string?
   !
   if ( strim == 0 .and. ptrim == 0 ) then
      match = .true.
   else
   !
   ! The string matches a literal part?
   !
      if ( ll > 0 ) then
         if ( string(start:min(strim,start+ll-1)) == literal(1:ll) ) then
            start = start + ll
            match = string_match( string(start:), pattern(p:) )
         endif
      endif
   endif
endif

if ( method == 1 ) then
   !
   ! Scan the whole of the remaining string ...
   !
   if ( ll == 0 ) then
      match = .true.
   else
      do while ( start <= strim )
         k = index( string(start:), literal(1:ll) )
         if ( k > 0 ) then
            start = start + k + ll - 1
            match = string_match( string(start:), pattern(p:) )
            if ( match ) then
               exit
            endif
         endif
         
         start = start + 1
      enddo
   endif
endif

if ( method == 2 .and. ll > 0 ) then
!
! Scan the whole of the remaining string ...
!
   if ( string(start:min(strim,start+ll-1)) == literal(1:ll) ) then
      match = string_match( string(start+ll:), pattern(p:) )
   endif
endif

return

end function string_match

!> if ( method == 2 .and. ll > 0 ) then
!> !
!> ! Scan the whole of the remaining string ...
!> !
!> if ( string(start:min(strim,start+ll-1)) == literal(1:ll) ) then
!> match = string_match( string(start+ll:), pattern(p:) )
!> endif
!> endif
!> return
!> end function string_match
!
!If method==2 and ll==0 then you should return .true.
!
! https://fortranwiki.org/fortran/show/match_wild

LOGICAL FUNCTION match_wild (pattern, string)
! compare given string for match to pattern which may
! contain wildcard characters:
! "?" matching any one character, and
! "*" matching any zero or more characters.
! Both strings may have trailing spaces which are ignored.
! Authors: Clive Page, userid: cgp  domain: le.ac.uk, 2003 (original code)
!          Rolf Sander, 2005 (bug fixes and pattern preprocessing)
! Minor bug fixed by Clive Page, 2005 Nov 29, bad comment fixed 2005 Dec 2.
! Serious bug fixed by  Robert H McClanahan, 2011 April 11th
!    This program is free software; you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation; either version 2 of the License, or
!    (at your option) any later version.
!
!    This program is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with this program; if not, write to the Free Software
!    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
!    02110-1301  USA
!
   IMPLICIT NONE

   CHARACTER(LEN=*), INTENT(IN) :: pattern ! pattern may contain * and ?
   CHARACTER(LEN=*), INTENT(IN) :: string  ! string to be compared
   INTEGER :: lenp, lenp2, lens, n, p2, p, s
   INTEGER :: n_question, n_asterisk
   LOGICAL :: found

   CHARACTER(LEN=LEN(pattern)) :: pattern2
   lens = LEN_TRIM(string)
   lenp = LEN_TRIM(pattern)

! If the pattern is empty, always return true
   IF (lenp == 0) THEN
     match_wild = .TRUE.
     RETURN
   ENDIF

! The pattern must be preprocessed. All consecutive occurrences of
! one or more question marks ('?') and asterisks ('*') are sorted and
! compressed. The result is stored in pattern2.

   pattern2(:)=''
   p  = 1 ! current position in pattern
   p2 = 1 ! current position in pattern2
   DO
     IF ((pattern(p:p) == '?').OR.(pattern(p:p) == '*')) THEN
! a special character was found in the pattern
       n_question = 0
       n_asterisk = 0
       DO WHILE (p <= lenp)
         ! count the consecutive question marks and asterisks
         IF ((pattern(p:p) /= '?').AND.(pattern(p:p) /= '*')) EXIT
         IF (pattern(p:p) == '?') n_question = n_question + 1
         IF (pattern(p:p) == '*') n_asterisk = n_asterisk + 1
         p = p + 1
       ENDDO
       IF (n_question>0) THEN ! first, all the question marks
         pattern2(p2:p2+n_question-1) = REPEAT('?',n_question)
         p2 = p2 + n_question
       ENDIF
       IF (n_asterisk>0) THEN ! next, the asterisk (only one!)
         pattern2(p2:p2) = '*'
         p2 = p2 + 1
       ENDIF
     ELSE
! just a normal character
       pattern2(p2:p2) = pattern(p:p)
       p2 = p2 + 1
       p = p + 1
     ENDIF
     IF (p > lenp) EXIT
   ENDDO
!!   lenp2 = p2 - 1
   lenp2 = len_trim(pattern2)

! The modified wildcard in pattern2 is compared to the string:

   p2 = 1
   s = 1
   match_wild = .FALSE.
   DO
     IF (pattern2(p2:p2) == '?') THEN
! accept any char in string
       p2 = p2 + 1
       s = s + 1

     ELSEIF (pattern2(p2:p2) == "*") THEN
       p2 = p2 + 1
       IF (p2 > lenp2) THEN
! anything goes in rest of string
         match_wild = .TRUE.
         EXIT ! .TRUE.
       ELSE
! search string for char at p2
         n = INDEX(string(s:), pattern2(p2:p2))
         IF (n == 0) EXIT  ! .FALSE.
         s = n + s - 1
       ENDIF
     ELSEIF (s.gt.len(string)) THEN  !JSU
        EXIT                         !JSU
     ELSEIF (pattern2(p2:p2) == string(s:s)) THEN
! single char match
       p2 = p2 + 1
       s = s + 1
     ELSE
! non-match
!       EXIT ! .FALSE.
! Previous line buggy because failure to match one character in the pattern
! does not mean that a match won't be found later. Back up through pattern string
! until first wildcard character is found and start over with the exact character
! match. If the end of the string is reached, then return .FALSE.
!      04/11/2011 Robert McClanahan    Robert.McClanahan   <<at>>   AECC.COM
!
       found = .FALSE.
       DO WHILE (p2 > 0 .AND. .NOT. found)
         p2 = p2 - 1
         IF (p2 == 0) EXIT  !  .FALSE.
         IF (pattern(p2:p2) == '*' .OR. pattern(p2:p2) == '?') found = .TRUE.
       END DO
       s = s + 1
     ENDIF

     IF (p2 > lenp2 .AND. s > lens) THEN
! end of both pattern2 and string
       match_wild = .TRUE.
       EXIT ! .TRUE.
     ENDIF

     IF (s > lens .AND. p2 == lenp) THEN
       IF(pattern2(p2:p2) == "*") THEN
! "*" at end of pattern2 represents an empty string
         match_wild = .TRUE.
         EXIT
       ENDIF
     ENDIF

     !JSU!IF (p2 > lenp2 .OR. s > lens ) THEN
     IF (p2 > lenp2 .OR. s > lens .or. p2 < 1) THEN
! end of either pattern2 or string
       EXIT ! .FALSE.
     ENDIF
   ENDDO

END FUNCTION match_wild

LOGICAL FUNCTION Wildcmp(pattern, candidate)

!Above cannot work for instance: match_wild('2','ABCDE2')
! Author: KunWing? with other c language contribution (by Jack Handy -
! <A href='mailto:jakkhandy@hotmail.com'>jakkhandy@hotmail.com</A>
! License: GPL license
! Revised on September 26, 2018 08:13:46 by Kun Wing? (165.225.64.66) (6706 characters / 2.0 pages)

CHARACTER(LEN=*), INTENT(IN) :: pattern ! pattern may contain * and ?
CHARACTER(LEN=*), INTENT(IN) :: candidate  ! string to be compared
Character(:),allocatable::pattStr,CandStr,tempstr   
integer::lenp,lenc,istar1,istar2,iquest,i,j

   i=scan(pattern,char(0))-1; j=scan(candidate,char(0))-1
   if(i==-1)i=len(pattern); if(j==-1)j=len(candidate)
   pattStr=trim(adjustl(pattern(:i))); CandStr=trim(adjustl(candidate(:j))); tempstr=CandStr
   lenp=len(pattStr); lenc=len(CandStr); i=1
   
   do while(pattStr(i:i)/='*')
      if(pattStr(i:i)/=CandStr(i:i).and.pattStr(i:i)/='?')then
         wildcmp=.false.; return; 
      endif
      i=i+1; if(i>=min(lenp,lenc))exit
   end do; wildcmp=(i==lenp).and.(lenp==lenc)
   
   j=i
   do while(j<=lenc)
      if(i<=lenp)then
         if(pattStr(i:i)=='*')then
            wildcmp=lenp==i; if(lenp==i)return
            i=i+1; !j=j+1
         elseif(i<=lenp.and.j<=lenc)then
            if(pattStr(i:i)==CandStr(j:j).or.pattStr(i:i)=='?')then
               i=i+1; j=j+1; !wildcmp=.true.
            else
               j=j+1; if(j>lenc)exit
            endif
         else
            j=j+1; 
            if(i>lenp.or.j>lenc)then
               wildcmp=.false.; return;
            endif 
         endif
      endif
   end do    

   wildcmp=(i-1==lenp)
   if(i>lenp)return

   do while(pattStr(i:i)=='*')
      i=i+1
   end do; 

   if(i<=lenp)wildcmp=(pattStr(i:i)=='*')

END FUNCTION Wildcmp

end module M_match
